#include "quantum_rng.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>

// Enhanced physical constants for quantum operations
#define QRNG_FINE_STRUCTURE 0x7297352743776A1BULL
#define QRNG_PLANCK 0x6955927086495225ULL
#define QRNG_RYDBERG 0x9E3779B97F4A7C15ULL
#define QRNG_ELECTRON_G 0x2B992DDFA232945ULL
#define QRNG_GOLDEN_RATIO 0x9E3779B97F4A7C15ULL
#define QRNG_SQRT2 0x6A09E667F3BCC908ULL

// New quantum mixing constants
#define QRNG_HEISENBERG 0xC13FA9A902A6328FULL
#define QRNG_SCHRODINGER 0x91E10DA5C79E7B1DULL
#define QRNG_PAULI_X 0x4C957F2D8A1E6B3CULL
#define QRNG_PAULI_Y 0xD3E99E3B6C1A4F78ULL
#define QRNG_PAULI_Z 0x8F142FC07892A5B6ULL

// Forward declarations of static functions
static inline double quantum_noise(double x);
static inline uint64_t splitmix64(uint64_t x);
static inline uint64_t hadamard_mix(uint64_t x);
static uint64_t get_system_entropy(void);
static uint64_t get_runtime_entropy(qrng_ctx *ctx);
static inline uint64_t hadamard_gate(uint64_t x);
static inline uint64_t phase_gate(uint64_t x, uint64_t angle);
static uint64_t measure_state(qrng_ctx *ctx, double quantum_state, uint64_t last);
static void quantum_step(qrng_ctx *ctx);

// Enhanced quantum noise function with multiple transformations
static inline double quantum_noise(double x) {
    volatile double noise = x;

    // Apply quantum uncertainty principle
    noise = sin(noise * M_PI) * cos(noise * M_E);
    noise = fabs(noise);

    // Apply Heisenberg uncertainty
    volatile double momentum = cos(noise * QRNG_FINE_STRUCTURE);
    volatile double position = sin(noise * QRNG_PLANCK);
    noise = (momentum * momentum + position * position) * 0.5;

    // Quantum tunneling effect
    noise = sqrt(noise * (1.0 - noise));

    // Normalize to [0,1]
    noise = noise - floor(noise);

    return noise;
}

// Enhanced splitmix64 with better avalanche
static inline uint64_t splitmix64(uint64_t x) {
    x ^= x >> 30;
    x *= 0xbf58476d1ce4e5b9ULL;
    x ^= x >> 27;
    x *= 0x94d049bb133111ebULL;
    x ^= x >> 31;
    x *= QRNG_HEISENBERG;
    x ^= x >> 29;
    return x;
}

// Enhanced Hadamard mixing function
static inline uint64_t hadamard_mix(uint64_t x) {
    x = splitmix64(x);
    x ^= QRNG_PAULI_X * (x >> 12);
    x *= QRNG_FINE_STRUCTURE;
    x ^= QRNG_PAULI_Y * (x >> 25);
    x *= QRNG_PLANCK;
    x ^= QRNG_PAULI_Z * (x >> 27);
    x *= QRNG_SCHRODINGER;
    x ^= x >> 13;
    return x;
}

// Enhanced system entropy collection
static uint64_t get_system_entropy(void) {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    uint64_t entropy = ((uint64_t)tv.tv_sec << 32) | tv.tv_usec;
    entropy ^= (uint64_t)getpid() << 32;
    entropy ^= (uint64_t)clock();

    // Use memory address of stack variable as additional entropy
    volatile void* stack_addr = &tv;
    entropy ^= (uint64_t)stack_addr;

    // Mix in CPU cycle count if available
    #if defined(__x86_64__) || defined(__i386__)
    unsigned int aux;
    entropy ^= __builtin_ia32_rdtsc();
    #endif

    return entropy;
}

// Enhanced runtime entropy collection
static uint64_t get_runtime_entropy(qrng_ctx *ctx) {
    struct timeval tv;
    gettimeofday(&tv, NULL);

    uint64_t runtime = ((uint64_t)tv.tv_sec << 32) | tv.tv_usec;
    runtime ^= ctx->system_entropy;
    runtime ^= ctx->unique_id;
    runtime ^= ctx->counter;
    runtime = hadamard_mix(runtime);

    return runtime;
}

// Enhanced Hadamard gate with improved quantum properties
static inline uint64_t hadamard_gate(uint64_t x) {
    volatile double state = (double)x / UINT64_MAX;
    state = quantum_noise(state);

    // Apply quantum superposition
    uint64_t superposition = (uint64_t)(state * UINT64_MAX) ^ x;
    superposition = hadamard_mix(superposition);

    // Apply phase rotation
    volatile double phase = quantum_noise(state + 0.5);
    uint64_t rotation = (uint64_t)(phase * UINT64_MAX);

    superposition ^= rotation;
    superposition = hadamard_mix(superposition);

    return superposition;
}

// Enhanced phase gate with improved entropy
static inline uint64_t phase_gate(uint64_t x, uint64_t angle) {
    volatile double phase = (double)angle / UINT64_MAX;
    phase = quantum_noise(phase);

    uint64_t mixed = (uint64_t)(phase * UINT64_MAX);
    mixed = hadamard_mix(mixed * QRNG_RYDBERG);

    // Apply quantum entanglement
    mixed ^= QRNG_PAULI_X * (mixed >> 17);
    mixed *= QRNG_HEISENBERG;
    mixed ^= QRNG_PAULI_Y * (mixed >> 23);
    mixed *= QRNG_SCHRODINGER;

    return x ^ mixed;
}

// Enhanced measurement function with improved entropy collection
static uint64_t measure_state(qrng_ctx *ctx, double quantum_state, uint64_t last) {
    // Update runtime entropy
    ctx->runtime_entropy = get_runtime_entropy(ctx);

    volatile double collapsed = quantum_noise(quantum_state +
        (double)ctx->runtime_entropy / UINT64_MAX);

    // Update entropy pool with runtime entropy
    ctx->entropy_pool[ctx->pool_index] = quantum_noise(
        ctx->entropy_pool[ctx->pool_index] + collapsed +
        (double)ctx->runtime_entropy / UINT64_MAX
    );
    ctx->pool_index = (ctx->pool_index + 1) & 15;

    // Mix entropy pool
    ctx->pool_mixer = hadamard_mix(ctx->pool_mixer ^
        (uint64_t)(ctx->entropy_pool[ctx->pool_index] * UINT64_MAX) ^
        ctx->runtime_entropy);

    uint64_t result = (uint64_t)(collapsed * UINT64_MAX);
    result = hadamard_mix(result ^ (last * QRNG_ELECTRON_G) ^ ctx->runtime_entropy);

    // Apply quantum gates with runtime entropy
    result ^= QRNG_PAULI_X * (ctx->pool_mixer >> 29);
    result *= QRNG_HEISENBERG;
    result ^= QRNG_PAULI_Y * (result >> 31);
    result *= QRNG_SCHRODINGER;
    result ^= QRNG_PAULI_Z * (result >> 27);

    return result;
}

// Core state management and output generation
static void quantum_step(qrng_ctx *ctx) {
    if (!ctx) return;

    ctx->counter++;
    uint64_t mixer = splitmix64(ctx->counter * QRNG_GOLDEN_RATIO);

    // Update runtime entropy
    ctx->runtime_entropy = get_runtime_entropy(ctx);

    // Enhanced mixing rounds with improved quantum gates
    for (int round = 0; round < QRNG_MIXING_ROUNDS; round++) {
        mixer = hadamard_mix(mixer ^ ctx->pool_mixer ^ ctx->runtime_entropy);

        for (int i = 0; i < QRNG_NUM_QUBITS; i++) {
            // Apply Hadamard gate with runtime entropy
            ctx->phase[i] = hadamard_gate(ctx->counter + mixer + i + round +
                ctx->runtime_entropy);

            // Update quantum state with runtime entropy
            ctx->quantum_state[i] = quantum_noise(
                (double)ctx->phase[i] / UINT64_MAX +
                ctx->entropy_pool[i & 15] +
                (double)ctx->runtime_entropy / UINT64_MAX
            );

            // Measure and entangle with runtime entropy
            uint64_t measured = measure_state(ctx, ctx->quantum_state[i],
                ctx->last_measurement[i]);
            ctx->entangle[i] = phase_gate(measured,
                ctx->counter ^ mixer ^ ctx->runtime_entropy);
            ctx->last_measurement[i] = measured;

            // Apply quantum entanglement between qubits
            if (i > 0) {
                ctx->entangle[i] ^= hadamard_mix(ctx->entangle[i-1] ^ mixer ^
                    ctx->runtime_entropy);
                ctx->quantum_state[i] = quantum_noise(
                    ctx->quantum_state[i] + ctx->quantum_state[i-1] +
                    (double)ctx->runtime_entropy / UINT64_MAX
                );
            }

            mixer = splitmix64(mixer ^ measured ^ ctx->pool_mixer ^
                ctx->runtime_entropy);
        }
    }

    // Fill output buffer with improved mixing
    uint64_t prev = mixer;
    for (size_t i = 0; i < QRNG_BUFFER_SIZE / sizeof(uint64_t); i++) {
        uint64_t current = measure_state(ctx,
            ctx->quantum_state[i % QRNG_NUM_QUBITS],
            ctx->entangle[i % QRNG_NUM_QUBITS]);

        // Enhanced output mixing with runtime entropy
        current = hadamard_mix(current ^ prev ^ ctx->pool_mixer ^
            ctx->runtime_entropy);
        current ^= QRNG_PAULI_X * (current >> 29);
        current *= QRNG_HEISENBERG;
        current ^= QRNG_PAULI_Y * (current >> 31);
        current *= QRNG_SCHRODINGER;

        ctx->buffer.words[i] = current;
        prev = current;
    }
    ctx->buffer_pos = 0;
}

// Public API implementations
qrng_error qrng_init(qrng_ctx **ctx, const uint8_t *seed, size_t seed_len) {
    if (!ctx) return QRNG_ERROR_NULL_CONTEXT;
    if (!seed && seed_len > 0) return QRNG_ERROR_NULL_BUFFER;

    *ctx = calloc(1, sizeof(qrng_ctx));
    if (!*ctx) return QRNG_ERROR_NULL_CONTEXT;

    // Initialize context with system-specific entropy
    gettimeofday(&(*ctx)->init_time, NULL);
    (*ctx)->pid = getpid();
    (*ctx)->system_entropy = get_system_entropy();
    (*ctx)->unique_id = splitmix64((*ctx)->system_entropy);
    (*ctx)->pool_mixer = QRNG_HEISENBERG ^ (*ctx)->unique_id;
    (*ctx)->runtime_entropy = get_runtime_entropy(*ctx);

    // Initialize entropy pool with multiple sources
    for (int i = 0; i < 16; i++) {
        (*ctx)->entropy_pool[i] = quantum_noise(
            (double)((*ctx)->system_entropy >> i) / UINT64_MAX +
            (double)((*ctx)->init_time.tv_usec >> (i % 20)) / UINT64_MAX +
            (double)((*ctx)->pid << (i % 16)) / UINT64_MAX +
            (double)(*ctx)->runtime_entropy / UINT64_MAX
        );
    }

    // Initialize quantum state with runtime entropy
    uint64_t mixer = QRNG_GOLDEN_RATIO ^ (*ctx)->system_entropy;
    for (size_t i = 0; i < QRNG_NUM_QUBITS; i++) {
        mixer = splitmix64(mixer ^
            (seed ? seed[i % seed_len] : 0) ^
            (*ctx)->runtime_entropy);

        (*ctx)->phase[i] = hadamard_gate(
            (seed ? seed[i % seed_len] : i) ^
            mixer ^ (*ctx)->unique_id ^
            (*ctx)->runtime_entropy
        );

        (*ctx)->quantum_state[i] = quantum_noise(
            (double)((*ctx)->phase[i] ^ (*ctx)->system_entropy) / UINT64_MAX +
            (*ctx)->entropy_pool[i % 16] +
            (double)(*ctx)->runtime_entropy / UINT64_MAX
        );

        (*ctx)->last_measurement[i] = measure_state(*ctx,
            (*ctx)->quantum_state[i],
            seed ? seed[(seed_len - 1 - i) % seed_len] : i);

        (*ctx)->entangle[i] = phase_gate((*ctx)->last_measurement[i],
            (seed ? seed[i % seed_len] : i) ^
            mixer ^ (*ctx)->runtime_entropy);
    }

    // Additional mixing rounds
    for (int i = 0; i < QRNG_MIXING_ROUNDS * 2; i++) {
        quantum_step(*ctx);
    }

    return QRNG_SUCCESS;
}

void qrng_free(qrng_ctx *ctx) {
    if (ctx) {
        memset(ctx, 0, sizeof(*ctx));
        free(ctx);
    }
}

qrng_error qrng_reseed(qrng_ctx *ctx, const uint8_t *seed, size_t seed_len) {
    if (!ctx) return QRNG_ERROR_NULL_CONTEXT;
    if (!seed && seed_len > 0) return QRNG_ERROR_NULL_BUFFER;
    if (seed_len == 0) return QRNG_ERROR_INVALID_LENGTH;

    // Update runtime entropy
    ctx->runtime_entropy = get_runtime_entropy(ctx);

    uint64_t mixer = QRNG_GOLDEN_RATIO ^ ctx->runtime_entropy;
    for (size_t i = 0; i < seed_len && i < QRNG_NUM_QUBITS; i++) {
        mixer = splitmix64(mixer ^ seed[i] ^ ctx->runtime_entropy);
        ctx->phase[i] = hadamard_gate(ctx->phase[i] ^ seed[i] ^ mixer ^
            ctx->runtime_entropy);
        ctx->quantum_state[i] = quantum_noise(
            (double)ctx->phase[i] / UINT64_MAX +
            (double)ctx->runtime_entropy / UINT64_MAX
        );
        ctx->last_measurement[i] = measure_state(ctx, ctx->quantum_state[i],
            seed[seed_len - 1 - i] ^ mixer);
        ctx->entangle[i] = phase_gate(ctx->last_measurement[i],
            seed[i] ^ mixer ^ ctx->runtime_entropy);
    }

    for (int i = 0; i < QRNG_MIXING_ROUNDS * 2; i++) {
        quantum_step(ctx);
    }

    return QRNG_SUCCESS;
}

qrng_error qrng_bytes(qrng_ctx *ctx, uint8_t *out, size_t len) {
    if (!ctx) return QRNG_ERROR_NULL_CONTEXT;
    if (!out) return QRNG_ERROR_NULL_BUFFER;
    if (len == 0) return QRNG_ERROR_INVALID_LENGTH;

    while (len > 0) {
        if (ctx->buffer_pos >= QRNG_BUFFER_SIZE) {
            quantum_step(ctx);
        }

        size_t copy_len = QRNG_BUFFER_SIZE - ctx->buffer_pos;
        if (copy_len > len) copy_len = len;

        memcpy(out, ctx->buffer.bytes + ctx->buffer_pos, copy_len);
        ctx->buffer_pos += copy_len;
        out += copy_len;
        len -= copy_len;
    }

    return QRNG_SUCCESS;
}

uint64_t qrng_uint64(qrng_ctx *ctx) {
    if (!ctx) return 0;

    uint64_t result;
    qrng_bytes(ctx, (uint8_t*)&result, sizeof(result));

    // Enhanced output mixing with runtime entropy
    ctx->runtime_entropy = get_runtime_entropy(ctx);
    result = splitmix64(result ^ ctx->runtime_entropy);
    result ^= QRNG_PAULI_X * (result >> 27);
    result *= QRNG_HEISENBERG;
    result ^= QRNG_PAULI_Y * (result >> 31);
    result *= QRNG_SCHRODINGER;
    result ^= QRNG_PAULI_Z * (result >> 29);

    return result;
}

double qrng_double(qrng_ctx *ctx) {
    if (!ctx) return 0.0;
    return (double)(qrng_uint64(ctx) >> 11) * (1.0/9007199254740992.0);
}

int32_t qrng_range32(qrng_ctx *ctx, int32_t min, int32_t max) {
    if (!ctx || min > max) {
        return max;
    }

    uint32_t range = (uint32_t)(max - min + 1);
    if (range == 0) {
        return max;
    }

    uint32_t threshold = (uint32_t)-range % range;
    uint32_t r;

    do {
        r = (uint32_t)qrng_uint64(ctx);
    } while (r < threshold);

    return min + (r % range);
}

uint64_t qrng_range64(qrng_ctx *ctx, uint64_t min, uint64_t max) {
    if (!ctx || min > max) {
        return max;
    }

    if (min == max) {
        return min;
    }

    uint64_t range = max - min + 1;
    if (range == 0) {
        return max;
    }

    uint64_t threshold = -range % range;
    uint64_t r;

    do {
        r = qrng_uint64(ctx);
    } while (r < threshold);

    return min + (r % range);
}

double qrng_get_entropy_estimate(qrng_ctx *ctx) {
    if (!ctx) return 0.0;

    // Calculate entropy estimate from the entropy pool
    double entropy = 0.0;
    for (int i = 0; i < 16; i++) {
        entropy += -log2(ctx->entropy_pool[i] + 1e-10);
    }

    // Include runtime entropy in the estimate
    ctx->runtime_entropy = get_runtime_entropy(ctx);
    entropy += -log2((double)(ctx->runtime_entropy & 0xFF) / 256.0 + 1e-10);

    return entropy / 17.0;  // Average over all sources
}

qrng_error qrng_entangle_states(qrng_ctx *ctx, uint8_t *state1, uint8_t *state2, size_t len) {
    if (!ctx) return QRNG_ERROR_NULL_CONTEXT;
    if (!state1 || !state2) return QRNG_ERROR_NULL_BUFFER;
    if (len == 0) return QRNG_ERROR_INVALID_LENGTH;

    // Update runtime entropy
    ctx->runtime_entropy = get_runtime_entropy(ctx);

    // Create entanglement between states using quantum gates
    uint64_t mixer = splitmix64(ctx->counter * QRNG_GOLDEN_RATIO);

    for (size_t i = 0; i < len; i++) {
        // Create superposition of states
        uint64_t s1 = hadamard_gate(state1[i] ^ mixer ^ ctx->runtime_entropy);
        uint64_t s2 = hadamard_gate(state2[i] ^ mixer ^ ctx->runtime_entropy);

        // Apply phase rotation to create correlation
        uint64_t phase = phase_gate(s1 ^ s2, ctx->counter ^ mixer ^ ctx->runtime_entropy);

        // Entangle the states
        state1[i] = (uint8_t)(s1 ^ phase);
        state2[i] = (uint8_t)(s2 ^ phase);

        // Update mixer for next iteration
        mixer = splitmix64(mixer ^ s1 ^ s2 ^ ctx->runtime_entropy);
    }

    // Update quantum state
    for (int i = 0; i < QRNG_NUM_QUBITS; i++) {
        ctx->quantum_state[i] = quantum_noise(
            ctx->quantum_state[i] +
            (double)ctx->runtime_entropy / UINT64_MAX
        );
    }

    return QRNG_SUCCESS;
}

qrng_error qrng_measure_state(qrng_ctx *ctx, uint8_t *state, size_t len) {
    if (!ctx) return QRNG_ERROR_NULL_CONTEXT;
    if (!state) return QRNG_ERROR_NULL_BUFFER;
    if (len == 0) return QRNG_ERROR_INVALID_LENGTH;

    // Update runtime entropy
    ctx->runtime_entropy = get_runtime_entropy(ctx);

    // Measure each byte of the state
    uint64_t mixer = splitmix64(ctx->counter * QRNG_GOLDEN_RATIO);

    for (size_t i = 0; i < len; i++) {
        // Create quantum state from byte
        double quantum_val = quantum_noise(
            (double)state[i] / 255.0 +
            (double)ctx->runtime_entropy / UINT64_MAX
        );

        // Perform measurement
        uint64_t measured = measure_state(ctx, quantum_val, mixer);

        // Collapse state to classical value
        state[i] = (uint8_t)(measured & 0xFF);

        // Update mixer for next iteration
        mixer = splitmix64(mixer ^ measured ^ ctx->runtime_entropy);
    }

    // Update quantum context state
    for (int i = 0; i < QRNG_NUM_QUBITS; i++) {
        ctx->last_measurement[i] = measure_state(ctx,
            ctx->quantum_state[i],
            ctx->last_measurement[i]);
    }

    return QRNG_SUCCESS;
}

const char* qrng_version(void) {
    static char version[32];
    snprintf(version, sizeof(version), "%d.%d.%d",
            QRNG_VERSION_MAJOR, QRNG_VERSION_MINOR, QRNG_VERSION_PATCH);
    return version;
}

const char* qrng_error_string(qrng_error err) {
    switch (err) {
        case QRNG_SUCCESS:
            return "Success";
        case QRNG_ERROR_NULL_CONTEXT:
            return "Null context error";
        case QRNG_ERROR_NULL_BUFFER:
            return "Null buffer error";
        case QRNG_ERROR_INVALID_LENGTH:
            return "Invalid length error";
        case QRNG_ERROR_INSUFFICIENT_ENTROPY:
            return "Insufficient entropy error";
        case QRNG_ERROR_INVALID_RANGE:
            return "Invalid range parameters";
        default:
            return "Unknown error";
    }
}
